#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-12-01
################################################################

import copy
from hex_zmq_servers import HexLaunch, HexNodeConfig
from hex_zmq_servers import HEX_ZMQ_SERVERS_PATH_DICT, HEX_ZMQ_CONFIGS_PATH_DICT
from importlib.util import find_spec

# YOCO config
YOCO = {
    "use_sim": True,
    "cam_type": "empty",
    "srv_port": {
        "mujoco_port": 12345,
        "robot_port": 12346,
        "camera_port": 12347,
    },
    "params": {
        "mujoco": {
            "headless": True,
        },
        "robot": {
            "mit_kp": [200.0, 200.0, 250.0, 150.0, 20.0, 20.0, 20.0],
            "mit_kd": [5.0, 5.0, 5.0, 5.0, 1.0, 1.0, 1.0],
            "arm_type": "archer_y6",
        },
        "rgb": {
            "resolution": [640, 480],
            "crop": [0, 640, 0, 480],
            "exposure": 70,
            "temperature": 0,
        },
        "realsense": {
            "resolution": [640, 480],
        },
        "berxel": {
            "exposure": 10000,
            "gain": 100,
        },
    },
    "device": {
        "robot": {
            "device_ip": "172.18.8.161",
            "device_port": 8439,
        },
        "camera": {}
    },
}

# Mujoco srv
MUJOCO_ARCHER_Y6_SRV = {
    "name": "mujoco_archer_y6_srv",
    "node_path": HEX_ZMQ_SERVERS_PATH_DICT["mujoco_archer_y6"],
    "cfg_path": HEX_ZMQ_CONFIGS_PATH_DICT["mujoco_archer_y6"],
    "cfg": {
        "net": {
            "ip": "127.0.0.1",
            "port": YOCO["srv_port"]["mujoco_port"],
        },
        "params": {
            "states_rate": 1000,
            "img_rate": 30,
            "tau_ctrl": False,
            "headless": YOCO["params"]["mujoco"]["headless"],
            "sens_ts": True,
            "mit_kp": YOCO["params"]["robot"]["mit_kp"],
            "mit_kd": YOCO["params"]["robot"]["mit_kd"],
            "cam_type": YOCO["cam_type"],
        },
    },
}

# Robot srv
ROBOT_HEXARM_SRV = {
    "name": "robot_archer_y6_srv",
    "node_path": HEX_ZMQ_SERVERS_PATH_DICT["robot_hexarm"],
    "cfg_path": HEX_ZMQ_CONFIGS_PATH_DICT["robot_hexarm"],
    "cfg": {
        "net": {
            "port": YOCO["srv_port"]["robot_port"],
        },
        "params": {
            "device_ip": YOCO["device"]["robot"]["device_ip"],
            "device_port": YOCO["device"]["robot"]["device_port"],
            "control_hz": 1000,
            "sens_ts": True,
            "arm_type": YOCO["params"]["robot"]["arm_type"],
            "use_gripper": True,
            "mit_kp": YOCO["params"]["robot"]["mit_kp"],
            "mit_kd": YOCO["params"]["robot"]["mit_kd"],
        },
    },
}

# RGB srv
RGB_SRV = {
    "name": "camera_archer_y6_srv",
    "node_path": HEX_ZMQ_SERVERS_PATH_DICT["cam_rgb"],
    "cfg_path": HEX_ZMQ_CONFIGS_PATH_DICT["cam_rgb"],
    "cfg": {
        "net": {
            "port": YOCO["srv_port"]["camera_port"],
        },
        "params": {
            "cam_path": "/dev/video0",
            "resolution": YOCO["params"]["rgb"]["resolution"],
            "crop": YOCO["params"]["rgb"]["crop"],
            "exposure": YOCO["params"]["rgb"]["exposure"],
            "temperature": YOCO["params"]["rgb"]["temperature"],
            "frame_rate": 30,
            "sens_ts": True,
        },
    },
}

# Realsense srv
_HAS_REALSENSE = find_spec("pyrealsense2") is not None
if _HAS_REALSENSE:
    REALSENSE_SRV = {
        "name": "camera_archer_y6_srv",
        "node_path": HEX_ZMQ_SERVERS_PATH_DICT["cam_realsense"],
        "cfg_path": HEX_ZMQ_CONFIGS_PATH_DICT["cam_realsense"],
        "cfg": {
            "net": {
                "port": YOCO["srv_port"]["camera_port"],
            },
            "params": {
                "serial_number": "243422073194",
                "resolution": YOCO["params"]["realsense"]["resolution"],
                "frame_rate": 30,
                "sens_ts": True,
            },
        },
    }

# Berxel srv
_HAS_BERXEL = find_spec("berxel_py_wrapper") is not None
if _HAS_BERXEL:
    BERXEL_SRV = {
        "name": "camera_archer_y6_srv",
        "node_path": HEX_ZMQ_SERVERS_PATH_DICT["cam_berxel"],
        "cfg_path": HEX_ZMQ_CONFIGS_PATH_DICT["cam_berxel"],
        "cfg": {
            "net": {
                "port": YOCO["srv_port"]["camera_port"],
            },
            "params": {
                "serial_number": "P100RYB4C03M2B322",
                "exposure": YOCO["params"]["berxel"]["exposure"],
                "gain": YOCO["params"]["berxel"]["gain"],
                "frame_rate": 30,
                "sens_ts": True,
            },
        },
    }


def set_mujoco_node_cfg(
    default_node_params_dict: dict,
    mujoco_params: dict,
    robot_params: dict,
):
    # mujoco
    default_node_params_dict["mujoco_archer_y6_srv"]["cfg"]["params"][
        "headless"] = mujoco_params.get(
            "headless",
            MUJOCO_ARCHER_Y6_SRV["cfg"]["params"]["headless"],
        )
    # robot
    default_node_params_dict["mujoco_archer_y6_srv"]["cfg"]["params"][
        "mit_kp"] = robot_params.get(
            "mit_kp",
            MUJOCO_ARCHER_Y6_SRV["cfg"]["params"]["mit_kp"],
        )
    default_node_params_dict["mujoco_archer_y6_srv"]["cfg"]["params"][
        "mit_kd"] = robot_params.get(
            "mit_kd",
            MUJOCO_ARCHER_Y6_SRV["cfg"]["params"]["mit_kd"],
        )
    return default_node_params_dict


def set_robot_node_cfg(
    default_node_params_dict: dict,
    robot_params: dict,
    robot_device: dict,
):
    # params
    default_node_params_dict["robot_archer_y6_srv"]["cfg"]["params"][
        "arm_type"] = robot_params.get(
            "arm_type",
            ROBOT_HEXARM_SRV["cfg"]["params"]["arm_type"],
        )
    default_node_params_dict["robot_archer_y6_srv"]["cfg"]["params"][
        "mit_kp"] = robot_params.get(
            "mit_kp",
            ROBOT_HEXARM_SRV["cfg"]["params"]["mit_kp"],
        )
    default_node_params_dict["robot_archer_y6_srv"]["cfg"]["params"][
        "mit_kd"] = robot_params.get(
            "mit_kd",
            ROBOT_HEXARM_SRV["cfg"]["params"]["mit_kd"],
        )
    # device
    default_node_params_dict["robot_archer_y6_srv"]["cfg"]["params"][
        "device_ip"] = robot_device.get(
            "device_ip",
            ROBOT_HEXARM_SRV["cfg"]["params"]["device_ip"],
        )
    default_node_params_dict["robot_archer_y6_srv"]["cfg"]["params"][
        "device_port"] = robot_device.get(
            "device_port",
            ROBOT_HEXARM_SRV["cfg"]["params"]["device_port"],
        )
    return default_node_params_dict


def set_rgb_node_cfg(
    default_node_params_dict: dict,
    rgb_params: dict,
    rgb_device: dict,
):
    # params
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "resolution"] = rgb_params.get(
            "resolution",
            RGB_SRV["cfg"]["params"]["resolution"],
        )
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "crop"] = rgb_params.get(
            "crop",
            RGB_SRV["cfg"]["params"]["crop"],
        )
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "exposure"] = rgb_params.get(
            "exposure",
            RGB_SRV["cfg"]["params"]["exposure"],
        )
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "temperature"] = rgb_params.get(
            "temperature",
            RGB_SRV["cfg"]["params"]["temperature"],
        )
    # device
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "cam_path"] = rgb_device.get(
            "cam_path",
            RGB_SRV["cfg"]["params"]["cam_path"],
        )
    return default_node_params_dict


def set_realsense_node_cfg(
    default_node_params_dict: dict,
    realsense_params: dict,
    realsense_device: dict,
):
    # params
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "resolution"] = realsense_params.get(
            "resolution",
            REALSENSE_SRV["cfg"]["params"]["resolution"],
        )
    # device
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "serial_number"] = realsense_device.get(
            "serial_number",
            REALSENSE_SRV["cfg"]["params"]["serial_number"],
        )
    return default_node_params_dict


def set_berxel_node_cfg(
    default_node_params_dict: dict,
    berxel_params: dict,
    berxel_device: dict,
):
    # params
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "exposure"] = berxel_params.get(
            "exposure",
            BERXEL_SRV["cfg"]["params"]["exposure"],
        )
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "gain"] = berxel_params.get(
            "gain",
            BERXEL_SRV["cfg"]["params"]["gain"],
        )
    # device
    default_node_params_dict["camera_archer_y6_srv"]["cfg"]["params"][
        "serial_number"] = berxel_device.get(
            "serial_number",
            BERXEL_SRV["cfg"]["params"]["serial_number"],
        )
    return default_node_params_dict


def get_node_cfgs(node_params_dict: dict = {}, launch_args: dict = YOCO):
    default_node_params_dict = {}
    use_sim = launch_args.get("use_sim", YOCO["use_sim"])
    cam_type = launch_args.get("cam_type", YOCO["cam_type"])
    srv_port = launch_args.get("srv_port", YOCO["srv_port"])
    params = launch_args.get("params", YOCO["params"])
    device = launch_args.get("device", YOCO["device"])
    if use_sim:
        default_node_params_dict["mujoco_archer_y6_srv"] = copy.deepcopy(
            MUJOCO_ARCHER_Y6_SRV)
        default_node_params_dict["mujoco_archer_y6_srv"]["cfg"]["params"][
            "cam_type"] = cam_type
        default_node_params_dict["mujoco_archer_y6_srv"]["cfg"]["net"][
            "port"] = srv_port.get("mujoco_port",
                                   YOCO["srv_port"]["mujoco_port"])
        default_node_params_dict = set_mujoco_node_cfg(
            default_node_params_dict,
            params.get("mujoco", YOCO["params"]["mujoco"]),
            params.get("robot", YOCO["params"]["robot"]),
        )
    else:
        default_node_params_dict["robot_archer_y6_srv"] = copy.deepcopy(
            ROBOT_HEXARM_SRV)
        default_node_params_dict["robot_archer_y6_srv"]["cfg"]["net"][
            "port"] = srv_port.get("robot_port",
                                   YOCO["srv_port"]["robot_port"])
        default_node_params_dict = set_robot_node_cfg(
            default_node_params_dict,
            params.get("robot", YOCO["params"]["robot"]),
            device.get("robot", YOCO["device"]["robot"]),
        )
        # cam_type: empty, rgb, realsense, berxel
        if cam_type == "rgb":
            default_node_params_dict["camera_archer_y6_srv"] = copy.deepcopy(
                RGB_SRV)
            default_node_params_dict["camera_archer_y6_srv"]["cfg"]["net"][
                "port"] = srv_port.get("camera_port",
                                       YOCO["srv_port"]["camera_port"])
            default_node_params_dict = set_rgb_node_cfg(
                default_node_params_dict,
                params.get("rgb", YOCO["params"]["rgb"]),
                device.get("camera", YOCO["device"]["camera"]),
            )
        elif cam_type == "realsense":
            if _HAS_REALSENSE:
                default_node_params_dict[
                    "camera_archer_y6_srv"] = copy.deepcopy(REALSENSE_SRV)
                default_node_params_dict["camera_archer_y6_srv"]["cfg"]["net"][
                    "port"] = srv_port.get("camera_port",
                                           YOCO["srv_port"]["camera_port"])
                default_node_params_dict = set_realsense_node_cfg(
                    default_node_params_dict,
                    params.get("realsense", YOCO["params"]["realsense"]),
                    device.get("camera", YOCO["device"]["camera"]),
                )
        elif cam_type == "berxel":
            if _HAS_BERXEL:
                default_node_params_dict[
                    "camera_archer_y6_srv"] = copy.deepcopy(BERXEL_SRV)
                default_node_params_dict["camera_archer_y6_srv"]["cfg"]["net"][
                    "port"] = srv_port.get("camera_port",
                                           YOCO["srv_port"]["camera_port"])
                default_node_params_dict = set_berxel_node_cfg(
                    default_node_params_dict,
                    params.get("camera", YOCO["params"]["berxel"]),
                    device.get("camera", YOCO["device"]["camera"]),
                )
        elif cam_type == "empty":
            pass
        else:
            raise ValueError(f"unknown camera type: {cam_type}")

    return HexNodeConfig.parse_node_params_dict(
        node_params_dict,
        default_node_params_dict,
    )


def main():
    node_cfgs = get_node_cfgs()
    launch = HexLaunch(node_cfgs)
    launch.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-11-20
################################################################

from .hex_yoco_e3_desktop import HexYocoE3Desktop
from .hex_yoco_archer_y6 import HexYocoArcherY6

import os

file_dir = os.path.dirname(os.path.abspath(__file__))
HEX_YOCO_DRIVER_PATH_DICT = {
    "archer_y6": f"{file_dir}/drivers/archer_y6_driver.py",
    "e3_desktop": f"{file_dir}/drivers/e3_desktop_driver.py",
}

__all__ = [
    "HexYocoE3Desktop",
    "HexYocoArcherY6",
    "HEX_YOCO_DRIVER_PATH_DICT",
]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-11-21
################################################################

import numpy as np
from hex_zmq_servers import HexMujocoArcherY6Client
from hex_zmq_servers import HexRobotHexarmClient
from hex_zmq_servers import HexCamRGBClient

from importlib.util import find_spec

_HAS_BERXEL = find_spec("berxel_py_wrapper") is not None
_HAS_REALSENSE = find_spec("pyrealsense2") is not None
if _HAS_BERXEL:
    from hex_zmq_servers import HexCamBerxelClient
if _HAS_REALSENSE:
    from hex_zmq_servers import HexCamRealsenseClient

CAMERA_CONFIG = {
    "empty": [(False, False), (None, None)],
    "rgb": [(True, False), ((480, 640, 3), None)],
    "berxel": [(True, True), ((400, 640, 3), (400, 640))],
    "realsense": [(True, True), ((480, 640, 3), (480, 640))],
}


class HexYocoArcherY6:

    def __init__(
        self,
        yoco_config: dict,
        net_config: dict,
        prefix: str = "",
    ):
        self.__prefix = prefix
        if self.__prefix != "":
            self.__prefix = f"{self.__prefix}_"
        try:
            use_sim = yoco_config["use_sim"]
            cam_type = yoco_config["cam_type"]
            if cam_type == "berxel" and not _HAS_BERXEL:
                print(
                    "`berxel_py_wrapper` not found, setting cam_type to empty")
                cam_type = "empty"
            elif cam_type == "realsense" and not _HAS_REALSENSE:
                print("`pyrealsense2` not found, setting cam_type to empty")
                cam_type = "empty"
            if use_sim:
                mujoco_net_config = net_config[f"{self.__prefix}mujoco_net"]
            else:
                robot_net_config = net_config[f"{self.__prefix}robot_net"]
                camera_net_config = net_config[
                    f"{self.__prefix}camera_net"] if cam_type != "empty" else None
        except KeyError as ke:
            missing_key = ke.args[0]
            raise ValueError(
                f"Missing key: [{missing_key}] in yoco_config or net_config")

        self.__use_sim = use_sim
        self.__cam_type = cam_type
        (self.__use_rgb,
         self.__use_depth), (self.__rgb_shape,
                             self.__depth_shape) = CAMERA_CONFIG.get(
                                 cam_type, [(False, False), (None, None)])

        self.__clients = {}
        if self.__use_sim:
            self.__clients["mujoco"] = HexMujocoArcherY6Client(
                net_config=mujoco_net_config,
                recv_config={
                    "rgb": self.__use_rgb,
                    "depth": self.__use_depth,
                    "obj": False,
                },
            )
        else:
            self.__clients["robot"] = HexRobotHexarmClient(
                net_config=robot_net_config)
            self.__clients["camera"] = None
            if cam_type == "berxel":
                self.__clients["camera"] = HexCamBerxelClient(
                    net_config=camera_net_config)
            elif cam_type == "realsense":
                self.__clients["camera"] = HexCamRealsenseClient(
                    net_config=camera_net_config)
            elif cam_type == "rgb":
                self.__clients["camera"] = HexCamRGBClient(
                    net_config=camera_net_config)

    def __del__(self):
        for client in self.__clients.values():
            if client is not None:
                client.close()

    def get_yoco_config(self):
        return {
            "use_sim": self.__use_sim,
            "cam_type": self.__cam_type,
        }

    def get_cam_state(self):
        return {
            "use_rgb": self.__use_rgb,
            "use_depth": self.__use_depth,
            "rgb_shape": self.__rgb_shape,
            "depth_shape": self.__depth_shape,
        }

    def is_working(self):
        if self.__use_sim:
            return self.__clients["mujoco"].is_working()
        else:
            robot_working = self.__clients["robot"].is_working()
            camera_working = self.__clients["camera"].is_working(
            ) if self.__clients["camera"] is not None else True
            return robot_working and camera_working

    def reset(self):
        if self.__use_sim:
            return self.__clients["mujoco"].reset()
        else:
            raise ValueError("`reset` is not supported in real mode")

    def seq_clear(self):
        if self.__use_sim:
            return self.__clients["mujoco"].seq_clear()
        else:
            return self.__clients["robot"].seq_clear()

    def get_dofs(self):
        if self.__use_sim:
            return self.__clients["mujoco"].get_dofs()[0]
        else:
            return self.__clients["robot"].get_dofs()[0]

    def get_limits(self):
        if self.__use_sim:
            return self.__clients["mujoco"].get_limits()[0].reshape(-1, 1, 2)
        else:
            return self.__clients["robot"].get_limits()[0]

    def get_states(self, newest: bool = False):
        if self.__use_sim:
            return self.__clients["mujoco"].get_states(newest=newest)
        else:
            return self.__clients["robot"].get_states(newest=newest)

    def set_cmds(self, cmds: np.ndarray) -> bool:
        if self.__use_sim:
            return self.__clients["mujoco"].set_cmds(cmds)
        else:
            return self.__clients["robot"].set_cmds(cmds)

    def get_intri(self):
        if self.__use_rgb or self.__use_depth:
            if self.__use_sim:
                _, intri_array = self.__clients["mujoco"].get_intri()
                return intri_array
            else:
                _, intri_array = self.__clients["camera"].get_intri()
                return intri_array
        else:
            raise ValueError(
                f"`get_intri` is not supported with type {self.__cam_type}")

    def get_rgb(self, newest: bool = False):
        if self.__use_rgb:
            if self.__use_sim:
                return self.__clients["mujoco"].get_rgb(newest=newest)
            else:
                return self.__clients["camera"].get_rgb(newest=newest)
        else:
            raise ValueError(
                f"`get_rgb` is not supported with type {self.__cam_type}")

    def get_depth(self, newest: bool = False):
        if self.__use_depth:
            if self.__use_sim:
                return self.__clients["mujoco"].get_depth(newest=newest)
            else:
                return self.__clients["camera"].get_depth(newest=newest)
        else:
            raise ValueError(
                f"`get_depth` is not supported with type {self.__cam_type}")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
################################################################
# Copyright 2025 Dong Zhaorui. All rights reserved.
# Author: Dong Zhaorui 847235539@qq.com
# Date  : 2025-11-21
################################################################

import numpy as np
from hex_zmq_servers import HexMujocoE3DesktopClient
from hex_zmq_servers import HexRobotHexarmClient
from hex_zmq_servers import HexCamRGBClient

from importlib.util import find_spec

_HAS_BERXEL = find_spec("berxel_py_wrapper") is not None
_HAS_REALSENSE = find_spec("pyrealsense2") is not None
if _HAS_BERXEL:
    from hex_zmq_servers import HexCamBerxelClient
if _HAS_REALSENSE:
    from hex_zmq_servers import HexCamRealsenseClient

CAMERA_CONFIG = {
    "empty": [(False, False), (None, None)],
    "rgb": [(True, False), ((480, 640, 3), None)],
    "berxel": [(True, True), ((400, 640, 3), (400, 640))],
    "realsense": [(True, True), ((480, 640, 3), (480, 640))],
}


class HexYocoE3Desktop:

    def __init__(
        self,
        yoco_config: dict,
        net_config: dict,
        prefix: str = "",
    ):
        self.__prefix = prefix
        if self.__prefix != "":
            self.__prefix = f"{self.__prefix}_"
        try:
            use_sim = yoco_config["use_sim"]
            cam_type = yoco_config["cam_type"]
            for idx in range(len(cam_type)):
                if cam_type[idx] == "berxel" and not _HAS_BERXEL:
                    print(
                        "`berxel_py_wrapper` not found, setting cam_type to empty"
                    )
                    cam_type[idx] = "empty"
                elif cam_type[idx] == "realsense" and not _HAS_REALSENSE:
                    print(
                        "`pyrealsense2` not found, setting cam_type to empty")
                    cam_type[idx] = "empty"
            if use_sim:
                mujoco_net_config = net_config[f"{self.__prefix}mujoco_net"]
            else:
                left_robot_net_config = net_config[
                    f"{self.__prefix}left_robot_net"]
                right_robot_net_config = net_config[
                    f"{self.__prefix}right_robot_net"]
                camera_net_config = {}
                for idx, cam_name in enumerate(["head", "left", "right"]):
                    camera_net_config[cam_name] = net_config[
                        f"{self.__prefix}{cam_name}_camera_net"] if cam_type[
                            idx] != "empty" else None
        except KeyError as ke:
            missing_key = ke.args[0]
            raise ValueError(
                f"Missing key: [{missing_key}] in yoco_config or net_config")

        self.__use_sim = use_sim
        self.__cam_type = cam_type
        self.__use_rgb, self.__use_depth, self.__rgb_shape, self.__depth_shape = {}, {}, {}, {}
        for idx, cam_name in enumerate(["head", "left", "right"]):
            (self.__use_rgb[cam_name], self.__use_depth[cam_name]), (
                self.__rgb_shape[cam_name],
                self.__depth_shape[cam_name]) = CAMERA_CONFIG.get(
                    cam_type[idx], [(False, False), (None, None)])

        self.__clients = {}
        if self.__use_sim:
            self.__clients["mujoco"] = HexMujocoE3DesktopClient(
                net_config=mujoco_net_config,
                recv_config={
                    "head_rgb": self.__use_rgb["head"],
                    "left_rgb": self.__use_rgb["left"],
                    "right_rgb": self.__use_rgb["right"],
                    "head_depth": self.__use_depth["head"],
                    "left_depth": self.__use_depth["left"],
                    "right_depth": self.__use_depth["right"],
                    "obj": False,
                },
            )
        else:
            self.__clients["left_robot"] = HexRobotHexarmClient(
                net_config=left_robot_net_config)
            self.__clients["right_robot"] = HexRobotHexarmClient(
                net_config=right_robot_net_config)
            for cam_name in ["head", "left", "right"]:
                self.__clients[cam_name] = None
                if cam_type[idx] == "berxel":
                    self.__clients[f"{cam_name}_camera"] = HexCamBerxelClient(
                        net_config=camera_net_config[cam_name])
                elif cam_type[idx] == "realsense":
                    self.__clients[
                        f"{cam_name}_camera"] = HexCamRealsenseClient(
                            net_config=camera_net_config[cam_name])
                elif cam_type[idx] == "rgb":
                    self.__clients[f"{cam_name}_camera"] = HexCamRGBClient(
                        net_config=camera_net_config[cam_name])

    def __del__(self):
        for client in self.__clients.values():
            if client is not None:
                client.close()

    def get_yoco_config(self):
        return {
            "use_sim": self.__use_sim,
            "cam_type": self.__cam_type,
        }

    def get_cam_state(self):
        return {
            "use_rgb": self.__use_rgb,
            "use_depth": self.__use_depth,
            "rgb_shape": self.__rgb_shape,
            "depth_shape": self.__depth_shape,
        }

    def is_working(self):
        if self.__use_sim:
            return self.__clients["mujoco"].is_working()
        else:
            left_working = self.__clients["left_robot"].is_working()
            right_working = self.__clients["right_robot"].is_working()
            camera_working = True
            for cam_name in ["head", "left", "right"]:
                camera_working = camera_working and (
                    self.__clients[cam_name].is_working()
                    if self.__clients[cam_name] is not None else True)
            return left_working and right_working and camera_working

    def reset(self):
        if self.__use_sim:
            return self.__clients["mujoco"].reset()
        else:
            raise ValueError("`reset` is not supported in real mode")

    def seq_clear(self):
        if self.__use_sim:
            clear_hdr = self.__clients["mujoco"].seq_clear()
            return {
                "left": clear_hdr,
                "right": clear_hdr,
            }
        else:
            return {
                "left": self.__clients["left_robot"].seq_clear(),
                "right": self.__clients["right_robot"].seq_clear(),
            }

    def get_dofs(self):
        if self.__use_sim:
            dofs_list = self.__clients["mujoco"].get_dofs()
            return {
                "left": dofs_list[0],
                "right": dofs_list[1],
            }
        else:
            return {
                "left": self.__clients["left_robot"].get_dofs()[0],
                "right": self.__clients["right_robot"].get_dofs()[0],
            }

    def get_limits(self):
        if self.__use_sim:
            limits_list = self.__clients["mujoco"].get_limits()
            return {
                "left": limits_list[0].reshape(-1, 1, 2),
                "right": limits_list[1].reshape(-1, 1, 2),
            }
        else:
            return {
                "left": self.__clients["left_robot"].get_limits()[0],
                "right": self.__clients["right_robot"].get_limits()[0],
            }

    def get_states(self, robot_name: str, newest: bool = False):
        if robot_name not in ["left", "right"]:
            raise ValueError(f"robot_name must be in ['left', 'right']")

        if self.__use_sim:
            return self.__clients["mujoco"].get_states(robot_name=robot_name,
                                                       newest=newest)
        else:
            robot_key = None
            if robot_name == "left":
                robot_key = "left_robot"
            elif robot_name == "right":
                robot_key = "right_robot"
            else:
                raise ValueError(f"Invalid robot name: [{robot_name}]")
            return self.__clients[robot_key].get_states(newest=newest)

    def set_cmds(self, cmds: np.ndarray, robot_name: str) -> bool:
        if robot_name not in ["left", "right"]:
            raise ValueError(f"robot_name must be in ['left', 'right']")

        if self.__use_sim:
            return self.__clients["mujoco"].set_cmds(cmds, robot_name)
        else:
            robot_key = None
            if robot_name == "left":
                robot_key = "left_robot"
            elif robot_name == "right":
                robot_key = "right_robot"
            else:
                raise ValueError(f"Invalid robot name: [{robot_name}]")
            return self.__clients[robot_key].set_cmds(cmds)

    def get_intri(self):
        use_cam = False
        for cam_name in ["head", "left", "right"]:
            use_cam = use_cam or self.__use_rgb[cam_name] or self.__use_depth[
                cam_name]
        if use_cam:
            if self.__use_sim:
                _, intri_array = self.__clients["mujoco"].get_intri()
                print(f"intri_array: {intri_array}")
                return {
                    "head": intri_array[0],
                    "left": intri_array[1],
                    "right": intri_array[2],
                }
            else:
                return {
                    "head":
                    self.__clients["head_camera"].get_intri()[1] if
                    self.__clients["head_camera"] is not None else np.zeros(4),
                    "left":
                    self.__clients["left_camera"].get_intri()[1] if
                    self.__clients["left_camera"] is not None else np.zeros(4),
                    "right":
                    self.__clients["right_camera"].get_intri()[1]
                    if self.__clients["right_camera"] is not None else
                    np.zeros(4),
                }
        else:
            raise ValueError(
                f"`get_intri` is not supported with type {self.__cam_type}")

    def get_rgb(self, camera_name: str, newest: bool = False):
        if camera_name not in ["head", "left", "right"]:
            raise ValueError(
                f"camera_name must be in ['head', 'left', 'right']")

        if self.__use_rgb[camera_name]:
            if self.__use_sim:
                return self.__clients["mujoco"].get_rgb(
                    camera_name=camera_name, newest=newest)
            else:
                return self.__clients[f"{camera_name}_camera"].get_rgb(
                    newest=newest)
        else:
            raise ValueError(
                f"`get_rgb` is not supported with type {self.__cam_type[camera_name]}"
            )

    def get_depth(self, camera_name: str, newest: bool = False):
        if camera_name not in ["head", "left", "right"]:
            raise ValueError(
                f"camera_name must be in ['head', 'left', 'right']")

        if self.__use_depth[camera_name]:
            if self.__use_sim:
                return self.__clients["mujoco"].get_depth(
                    camera_name=camera_name, newest=newest)
            else:
                return self.__clients[f"{camera_name}_camera"].get_depth(
                    newest=newest)
        else:
            raise ValueError(
                f"`get_depth` is not supported with type {self.__cam_type[camera_name]}"
            )
